﻿using HeuristicLab.Core;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System.Collections.Generic;
using HeuristicLab.Common;
using System.Text;
using System.Globalization;
using System.Linq;
using System;
using HEAL.Attic;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Formatters { 
  [StorableType("54D917E8-134E-4066-9A60-2737C12D81DC")]
  [Item("Infix String Formater", "Formatter for symbolic expressions, which produces an infix expression " +
    "as well as a list of all coefficient values")]
  public sealed class InfixExpressionStringFormatter : NamedItem, ISymbolicExpressionTreeStringFormatter {
    private readonly List<KeyValuePair<string, double>> constants;
    private int constantCounter = 0;//counter

    [StorableConstructor]
    private InfixExpressionStringFormatter(StorableConstructorFlag _): base(_) {}
    private InfixExpressionStringFormatter(InfixExpressionStringFormatter original, Cloner cloner) 
      : base(original, cloner) {
      constants = new List<KeyValuePair<string, double>>(original.constants);
      constantCounter = original.constantCounter;
    }
    public InfixExpressionStringFormatter() : base() {
      Name = ItemName;
      Description = ItemDescription;
      constants = new List<KeyValuePair<string, double>>();
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new InfixExpressionStringFormatter(this, cloner);
    }

    public string Format(ISymbolicExpressionTree symbolicExpressionTree) {
      StringBuilder strBuilder = new StringBuilder();
      constants.Clear();
      constantCounter = 0;
      FormatRecursively(symbolicExpressionTree.Root.GetSubtree(0).GetSubtree(0), strBuilder);
      strBuilder.Append($"{Environment.NewLine}{Environment.NewLine}");

      int maxDigits = GetDigits(constantCounter);
      int padding = constants.Max(x => x.Value.ToString("F12", CultureInfo.InvariantCulture).Length);
      foreach (var constant in constants) {
        int digits = GetDigits(Int32.Parse(constant.Key.Substring(2)));
        strBuilder.Append($"{constant.Key}{new String(' ', maxDigits - digits)} = " +
          string.Format($"{{0,{padding}:F12}}", constant.Value, CultureInfo.InvariantCulture) + Environment.NewLine);
      }
      return strBuilder.ToString();
    }

    private void FormatRecursively(ISymbolicExpressionTreeNode node, StringBuilder strBuilder) {
      if (node.SubtreeCount > 1) {
        var token = GetToken(node.Symbol);
        if (token == "+" || token == "-" || token == "OR" || token == "XOR") {
          strBuilder.Append("(");
          FormatRecursively(node.Subtrees.First(), strBuilder);

          foreach (var subtree in node.Subtrees.Skip(1)) {
            strBuilder.Append(" ").Append(token).Append(" ");
            FormatRecursively(subtree, strBuilder);
          }
          strBuilder.Append(")");

        } else if (token == "*" || token == "/" || token == "AND") {
          strBuilder.Append("(");
          FormatRecursively(node.Subtrees.First(), strBuilder);

          foreach (var subtree in node.Subtrees.Skip(1)) {
            strBuilder.Append(" ").Append(token).Append(" ");
            FormatRecursively(subtree, strBuilder);
          }
          strBuilder.Append(")");
        } else {
          // function with multiple arguments
          strBuilder.Append(token).Append("(");
          FormatRecursively(node.Subtrees.First(), strBuilder);
          foreach (var subtree in node.Subtrees.Skip(1)) {
            strBuilder.Append(", ");
            FormatRecursively(subtree, strBuilder);
          }
          strBuilder.Append(")");
        }
      } else if (node.SubtreeCount == 1) {
        var token = GetToken(node.Symbol);
        if (token == "-" || token == "NOT") {
          strBuilder.Append("(").Append(token).Append("(");
          FormatRecursively(node.GetSubtree(0), strBuilder);
          strBuilder.Append("))");
        } else if (token == "/") {
          strBuilder.Append("1/");
          FormatRecursively(node.GetSubtree(0), strBuilder);
        } else if (token == "+" || token == "*") {
          FormatRecursively(node.GetSubtree(0), strBuilder);
        } else {
          // function with only one argument
          strBuilder.Append(token).Append("(");
          FormatRecursively(node.GetSubtree(0), strBuilder);
          strBuilder.Append(")");
        }
      } else {
        if (node.Symbol is LaggedVariable) {
          var varNode = node as LaggedVariableTreeNode;
          if (!varNode.Weight.IsAlmost(1.0)) {
            strBuilder.Append("(");
            strBuilder.AppendFormat(CultureInfo.InvariantCulture, "{0}", varNode.Weight);
            strBuilder.Append("*");
          }
          strBuilder.Append("LAG(");
          if (varNode.VariableName.Contains("'")) {
            strBuilder.AppendFormat("\"{0}\"", varNode.VariableName);
          } else {
            strBuilder.AppendFormat("'{0}'", varNode.VariableName);
          }
          strBuilder.Append(", ")
            .AppendFormat(CultureInfo.InvariantCulture, "{0}", varNode.Lag)
            .Append(")");
        } else if (node.Symbol is Variable) {
          var varNode = node as VariableTreeNode;
          if (!varNode.Weight.IsAlmost(1.0)) {
            string constantKey = $"c_{constantCounter}";
            strBuilder.Append("(");
            strBuilder.AppendFormat(CultureInfo.InvariantCulture, "{0}", constantKey);
            strBuilder.Append("*");
            constants.Add(new KeyValuePair<string, double>(constantKey, varNode.Weight));
            constantCounter++;
          }
          if (varNode.VariableName.Contains("'")) {
            strBuilder.AppendFormat("\"{0}\"", varNode.VariableName);
          } else {
            strBuilder.AppendFormat("'{0}'", varNode.VariableName);
          }
          if (!varNode.Weight.IsAlmost(1.0)) {
            strBuilder.Append(")");
          }
        } else if (node.Symbol is FactorVariable) {
          var factorNode = node as FactorVariableTreeNode;
          if (factorNode.VariableName.Contains("'")) {
            strBuilder.AppendFormat("\"{0}\"", factorNode.VariableName);
          } else {
            strBuilder.AppendFormat("'{0}'", factorNode.VariableName);
          }
          strBuilder.AppendFormat("[{0}]",
            string.Join(", ", factorNode.Weights.Select(w => w.ToString(CultureInfo.InvariantCulture))));
        } else if (node.Symbol is BinaryFactorVariable) {
          var factorNode = node as BinaryFactorVariableTreeNode;
          if (!factorNode.Weight.IsAlmost(1.0)) {
            strBuilder.Append("(");
            strBuilder.AppendFormat(CultureInfo.InvariantCulture, "{0}", factorNode.Weight);
            strBuilder.Append("*");
          }
          if (factorNode.VariableName.Contains("'")) {
            strBuilder.AppendFormat("\"{0}\"", factorNode.VariableName);
          } else {
            strBuilder.AppendFormat("'{0}'", factorNode.VariableName);
          }
          strBuilder.Append(" = ");
          if (factorNode.VariableValue.Contains("'")) {
            strBuilder.AppendFormat("\"{0}\"", factorNode.VariableValue);
          } else {
            strBuilder.AppendFormat("'{0}'", factorNode.VariableValue);
          }

          if (!factorNode.Weight.IsAlmost(1.0)) {
            strBuilder.Append(")");
          }

        } else if (node.Symbol is Constant) {
          var constNode = node as ConstantTreeNode;
          string constantKey = $"c_{constantCounter}";

          strBuilder.AppendFormat(CultureInfo.InvariantCulture, constantKey);
          constants.Add(new KeyValuePair<string, double>(constantKey, constNode.Value));
          constantCounter++;
        }
      }
    }

    private int GetDigits(int x) {
      if (x == 0) return 1;
      return (int) Math.Floor(Math.Log10(x) + 1);
    }

    private string GetToken(ISymbol symbol) {
      var tok = InfixExpressionParser.knownSymbols.GetBySecond(symbol).SingleOrDefault();
      if (tok == null)
        throw new ArgumentException(string.Format("Unknown symbol {0} found.", symbol.Name));
      return tok;
    }
  }
}
